(defstruct path
  (cost 0 :type fixnum)
  (cost-estimate 0 :type fixnum)
  reversed
  (x 0 :type fixnum)
  (y 0 :type fixnum))

(declaim (inline manhattan-heuristic find-path-common))

;; Our heuristic metric. In fact, if we didn't have extra penalties
;; around water, this would be exact (well, almost).
(defun manhattan-heuristic (here-x here-y end-x end-y)
  (declare (fixnum here-x here-y end-x end-y))
  (+ (abs (- here-x end-x))
     (abs (- here-y end-y))))

;; Small-scale searching
;; If the path goes too far from current location, terminate
;; the search
(defun find-path-localized (start-x start-y end-x end-y)
  (declare (fixnum start-x start-y end-x end-y))
  (let ((visited (make-array (list (+ (* +lookout-window+ 2) 1)
                                   (+ (* +lookout-window+ 2) 1))
                             :element-type '(unsigned-byte 8))))
    (find-path-common start-x start-y end-x end-y
                      visited
                      (lambda (node visited end-x end-y)
                        (visit-nodes-local node visited start-x start-y
                                           end-x end-y)))))

;; Large scale searching, searches the whole map
(defun find-path (start-x start-y end-x end-y)
  (declare (fixnum start-x start-y end-x end-y))
  (let ((visited (make-array (list (+ *world-width* 2)
                                   (+ *world-height* 2))
                             :element-type '(unsigned-byte 8))))
    (find-path-common start-x start-y end-x end-y
                      visited
                      #'visit-nodes)))

;; The common code.
(defun find-path-common (start-x start-y end-x end-y visited visit-fn)
  (declare (fixnum start-x start-y end-x end-y))
  (let ((paths (list (make-path :cost 0 :reversed nil
                                :x start-x :y start-y
                                :cost-estimate
                                (manhattan-heuristic start-x start-y
                                                     end-x end-y)))))
    (loop for node = (car paths)
          for next = (funcall visit-fn node visited end-x end-y)
          while node
          if (and (= (path-x node) end-x)
                  (= (path-y node) end-y))
          return node
          else do
          (setf paths (merge 'list (cdr paths) next #'<
                             :key #'path-cost-estimate)))))

;; Ugly example of copy-paste code
;; This expands the nodes during search.
(defun visit-nodes (node visited end-x end-y)
  (declare (fixnum end-x end-y))
  (flet ((visit (i j direction)
           (when (zerop (aref visited i j))
             (setf (aref visited i j) 1)
             (let ((terrain (aref *world-map* i j)))
               (when (or (eql terrain #\.)
                         (eql terrain #\,))
                 (let ((cost (+ (path-cost node)
                                (if (eql terrain #\.) 1 3))))
                   (make-path :cost cost
                              :cost-estimate (+ cost
                                                (manhattan-heuristic i j
                                                                     end-x
                                                                     end-y))
                              :reversed (cons direction
                                              (path-reversed node))
                              :x i
                              :y j)))))))
    (when node
      (let* ((x (path-x node))
             (y (path-y node))
             (e (visit (1+ x) y :e))
             (w (visit (1- x) y :w))
             (n (visit x (1+ y) :n))
             (s (visit x (1- y) :s))
             (output '()))
        (when n
          (push n output))
        (when s
          (push s output))
        (when e
          (push e output))
        (when w
          (push w output))
        (sort output #'< :key #'path-cost-estimate)))))

(defun visit-nodes-local (node visited start-x start-y end-x end-y)
  (declare (fixnum start-x start-y end-x end-y))
  (flet ((visit (i j direction)
           (when (or (> (abs (- i start-x)) +lookout-window+)
                     (> (abs (- j start-y)) +lookout-window+))
             (return-from visit nil))
           (when (zerop (aref visited i j))
             (setf (aref visited i j) 1)
             (let ((terrain (aref *world-map* i j)))
               (when (or (eql terrain #\.)
                         (eql terrain #\,))
                 (let ((cost (+ (path-cost node)
                                (if (eql terrain #\.) 1 3))))
                   (make-path :cost cost
                              :cost-estimate (+ cost
                                                (manhattan-heuristic i j
                                                                     end-x
                                                                     end-y))
                              :reversed (cons direction
                                              (path-reversed node))
                              :x i
                              :y j)))))))
    (when node
      (let* ((x (path-x node))
             (y (path-y node))
             (e (visit (1+ x) y :e))
             (w (visit (1- x) y :w))
             (n (visit x (1+ y) :n))
             (s (visit x (1- y) :s))
             (output '()))
        (when n
          (push n output))
        (when s
          (push s output))
        (when e
          (push e output))
        (when w
          (push w output))
        (sort output #'< :key #'path-cost-estimate)))))