;;; Trivial readers
(defun collect-integers (line)
  (loop with pos = 0
        with value = 0
        do (multiple-value-setq (value pos)
             (parse-integer line :start pos :junk-allowed t))
        while value
        collect value))

;; An curious hack. We turn the "# " combination into a reader
;; macro, which can take a numeric argument.
;; Hence "#1235 " becomes legal input
;; To make this work 100% we need to tack a space at the end of
;; every input line, so input like "#3" would work.

(defun our-sharp-space (stream subchar count)
  (declare (ignore subchar stream))
  (list :player count))

(let ((readtable (copy-readtable)))
  (set-dispatch-macro-character #\# #\Space #'our-sharp-space readtable)
  (defun read-server-response ()
    (let ((*package* (find-package "KEYWORD"))
          (*readtable* readtable)
          (line (read-line *standard-input* nil nil)))

      (when (null line)
        (write-line "EOF on socket, exiting" *debug-io*)
        (ext:quit))

      ;; A quick hack to allow the #\Space reader to work
      ;; under all legal conditions.
      (setf line (concatenate 'string line " "))

      (loop with pos = 0
            with obj = nil
            do (multiple-value-setq (obj pos)
                 (read-from-string line nil nil :start pos))
            while obj
            collect obj))))

;;; Used after read-server-response to chop it into
;;; ((id actions...)
;;;  (id actions...)
(defun chop-server-response (response)
  ;; Chop off the first nil
  (cdr
   ;; The collector returns them in a wrong order, fix that
   (nreverse
    (loop with collector
          with uber-list = nil

          for x in response

          if (consp x)
          do (progn
               (push (nreverse collector) uber-list)
               (setf collector nil)
               ;; Push player id
               (push (second x) collector))

          else do (push x collector)

          finally
          (push (nreverse collector) uber-list)
          (return uber-list)))))

;;; Initialization readers
;;; Setup  certain global variables according
;;; to the configuration
(defun read-configuration ()
  (destructuring-bind (id capacity money) (collect-integers (read-line))
    (setf *own-id* id
          *carry-capacity* capacity
          *initial-capital* money
          *capital* money)))

(defun read-list-of-packages ()
  (let ((integers (collect-integers (read-line))))
    (loop for (id x y weight . rest) on integers by #'cddddr
          collect (list id x y weight))))

;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;
;;; THE BOARD
;;;
(defun read-board ()
  ;;; Read board
  ;;; 1) Dimensions
  (destructuring-bind (width height) (collect-integers (read-line))
    (let* ((map (loop for i from 1 to height
                      collect (concatenate 'string "#" (read-line) "#")))

           (padded-map (nconc (list (make-array (+ width 2)
                                                :element-type 'base-char
                                                :initial-element #\#))
                              map
                              (list (make-array (+ width 2)
                                                :element-type 'base-char
                                                :initial-element #\#))))

           (flipped-map (make-array (list (+ height 2) (+ width 2))
                                    :element-type 'base-char
                                    :initial-contents padded-map)))

      ;; Flip the map so we can use x/y indexing on it
      (setf *world-map* (make-array (list (+ width 2) (+ height 2))
                                    :element-type 'base-char))
      (setf *world-height* height
            *world-width* width)

      (dotimes (x (+ width 2))
        (dotimes (y (+ height 2))
          do (setf (aref *world-map* x y)
                   (aref flipped-map y x))))

      ;; Another loop to do cleanup
      (loop for x from 1 to width do
            (loop for y from 1 to height
                  do (cond ((eql (aref *world-map* x y) #\~)
                            (increase-weight-around x y))

                           ((eql (aref *world-map* x y) #\@)
                            (push (make-homebase :x x :y y)
                                  *homebases*)
                            (setf (aref *world-map* x y) #\.))))))))

(defun increase-weight-around (x y)
  (flet ((test (i j)
           (when (and (<= 1 i *world-width*)
                      (<= 1 j *world-height*))
             (when (eql (aref *world-map* i j) #\.)
               (setf (aref *world-map* i j) #\,)))))
    (test (1- x) y)
    (test (1+ x) y)
    (test x (1- y))
    (test x (1+ y))))

;;; Sending things
;;; Commands
;;; Expects a symbol for direction, ie :n
(defun send-move (bid direction)
  (with-standard-io-syntax
    (decf *capital* (abs bid))
    (debug-print :move bid direction)
    (format t "~D Move ~A~%" bid direction)
    (force-output)))

;; Packages can be a list
(defun send-pick (bid packages)
  (with-standard-io-syntax
    (decf *capital* (abs bid))
    (debug-print :pick bid packages)
    (format t "~D Pick ~{~D~^ ~}~%" bid packages)
    (force-output)))

;; Likewise
(defun send-drop (bid packages)
  (with-standard-io-syntax
    (decf *capital* (abs bid))
    (debug-print :drop bid packages)
    (format t "~D Drop ~{~D~^ ~}~%" bid packages)
    (force-output)))

