(defvar *live-robots* nil)

(defun clean-dead-robots (response)
  (let* ((live-bots (mapcar #'car response))
         (dead-bot-ids (nset-difference *live-robots* live-bots))
         (dead-bots (mapcar (lambda (id) (robot-by-id id))
                            dead-bot-ids)))
    (dolist (bot dead-bots)
      ;; Remove them from bots hash
      (remhash (robot-id bot) *robots*)
      (when (= (robot-id bot) *own-id*)
        (debug-print "We are no longer alive")
        (exit-the-system))
      (dolist (pkg (robot-packages bot))
        (remhash (game-package-id pkg) *packages*)))

    (setf *live-robots* live-bots)))

(defun handle-server-response (response)
  (clean-dead-robots response)
  (dolist (data response)
    (let ((id (car data))
          (events (cdr data))
          (check-for-possible-delivery '()))

      (when (robot-by-id id)
        (setf (robot-events (robot-by-id id)) events))

      (debug-print :robot id :server-event events)
      (loop with robot = (robot-by-id id)
            for event = (pop events)
            while event do
            (case event
              (:n (incf (robot-y robot)))
              (:s (decf (robot-y robot)))
              (:e (incf (robot-x robot)))
              (:w (decf (robot-x robot)))

              ;; A new robot appeared, add it to the
              ;; robots hash
              (:x
               (let ((x (pop events))
                     ;; The first pop eats away :y
                     (y (progn (pop events) (pop events))))
                 (setf robot
                       (make-robot :id id :x x :y y)
                       (robot-by-id id) robot)))

              (:p
               (let* ((pkg-id (pop events))
                      (pkg (ensure-package-at
                            pkg-id
                            (robot-x robot)
                            (robot-y robot))))
                 ;; Set the package to the list of carried packages
                 (setf (game-package-in-motion pkg) t)
                 (push pkg (robot-packages robot))
                 ;; Clear the packages-at-hash
                 (setf (packages-at (robot-x robot) (robot-y robot))
                       (delete pkg (packages-at (robot-x robot)
                                                (robot-y robot))))))

              (:d
               (let* ((pkg-id (pop events))
                      (pkg (ensure-package-at
                            pkg-id
                            (robot-x robot)
                            (robot-y robot))))
                 ;; Move package here
                 (setf (game-package-at-x pkg) (robot-x robot)
                       (game-package-at-y pkg) (robot-y robot)
                       (game-package-in-motion pkg) nil)

                 ;; Remove from carry list
                 (setf (robot-packages robot)
                       (delete pkg (robot-packages robot)))

                 ;; We know the package was delivered, get
                 ;; rid of it
                 ;; Otherwise add it to location->package-list, too
                 (if (and (eql (game-package-at-x pkg)
                               (game-package-to-x pkg))
                          (eql (game-package-at-x pkg)
                               (game-package-to-x pkg)))
                     (remhash pkg-id *packages*)
                     (progn
                       (push pkg (packages-at (robot-x robot)
                                              (robot-y robot)))
                       ;; Add this package to be checked
                       ;; afterwards for interest factor
                       ;; adjustment
                       (push pkg check-for-possible-delivery)))))))

      ;; All the moves have been done, adjust interest factor
      ;; If there is another bot nearby, it was probably dropped
      ;; => multiply it by 0.95
      ;; Rationale: If the package gets dropped a lot, there's
      ;; a push-war going on
      ;; Otherwise it was probably delivered (or the other bot
      ;; is ST00P1D) and we set its value to a lowly 0.01.
      (let ((bot-locations (bot-locations)))
        (dolist (pkg check-for-possible-delivery)
          (if (find (list (game-package-at-x pkg)
                          (game-package-at-y pkg))
                    bot-locations
                    :test #'equal)
            (setf (game-package-interest-factor pkg)
                  (* 0.90 (game-package-interest-factor pkg)))
            (setf (game-package-interest-factor pkg) 0.01)))))))

;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;
;;; PACKAGE HANDLING

(defun ensure-package-at (id x y)
  (if (package id)
      (let ((package (package id)))
        (setf (game-package-at-x package) x
              (game-package-at-y package) y))
      (progn
        (setf (package id)
              (make-game-package :id id :at-x x :at-y y))
        (push (package id) (packages-at x y))))
  (gethash id *packages*))

(defun handle-package-list (package-list)
  (debug-print :package-list package-list)
  ;; Find our location
  (let ((our-x (our-bot-x))
        (our-y (our-bot-y)))

    ;; Make sure the packages do exist in the hash table
    (dolist (p package-list)
      (ensure-package-at (first p) our-x our-y)

      ;; Update the package stats
      (let ((package (package (first p))))
        (setf (game-package-to-x package) (second p)
              (game-package-to-y package) (third p)
              (game-package-weight package) (fourth p))

        ;;; FIXME: Tt ei vlttmtt haluta tehd
        ;;; ihan joka vitun paketille. Jos paketteja
        ;;; IHAN SAATANASTI, niin sit tulee JUNSTK
        ;;; Sdetn tllanen detaili maanantaina
        ;;; Mahdollisesti painon mukaan
        (when (null (game-package-undeliverable package))
          ;; If this is null, this is probably the first time
          ;; we've seen this
          ;; Update estimated package stats accordingly
          (update-package-stats (game-package-weight package))
          (if (reachable (game-package-to-x package)
                         (game-package-to-y package))
              (setf (game-package-undeliverable package) t)
              (setf (game-package-undeliverable package) :undeliverable)))))

    ;; Check if this is a homebase, and if so, update
    ;; homebase goodness estimate
    ;; Also updates the last-visited status
    (maybe-update-homebase-stats our-x our-y)

    ;; Find out and clean any old packages at this location
    (clean-old-packages our-x our-y package-list)

    ;; Mark a possible homebase as visited lately
    ;; Finally, return the size of the smallest package at this
    ;; location, or nil
    (let ((sizes-of-packages-here
           (mapcar #'game-package-weight
                   (remove :undeliverable (packages-at our-x our-y)
                           :key #'game-package-undeliverable))))
      (and sizes-of-packages-here
           (apply #'min sizes-of-packages-here)))))


(defun clean-old-packages (x y package-list)
  (let* ((existing-package-ids (mapcar #'car package-list))
         (existing-packages (mapcar (lambda (x) (package x))
                                    existing-package-ids))
         (supposed-packages (packages-at x y))
         (not-existing (nset-difference supposed-packages
                                        existing-packages)))

    (dolist (pkg not-existing)
      (remhash (game-package-id pkg) *packages*)
      (setf (packages-at x y) existing-packages))))
