;;; Basic strategies 0 and 1
;;; Strategy 0:
;;; If at place where one can drop packages, drop 'em
(declaim (inline find-packages-destined-here drop-packages
                 carrying-at-the-moment))

;; Any packages for current location?
(defun find-packages-destined-here ()
  (let ((x (our-bot-x))
        (y (our-bot-y))
        (pkgs (robot-packages (robot-by-id *own-id*))))
    (remove-if-not (lambda (pkg) (and (eql (game-package-to-x pkg) x)
                                      (eql (game-package-to-y pkg) y)))
                   pkgs)))

;; Send a drop command
(defun drop-packages (packages bid)
  (let ((ids (mapcar #'game-package-id packages)))
    (send-drop bid ids)))

;; What are we (or someone else) carrying right now?
(defun carrying-at-the-moment (&optional robot)
  (unless robot
    (setf robot (robot-by-id *own-id*)))
  (let ((packages (robot-packages robot)))
    (reduce #'+ packages :key (lambda (pkg)
                                (or (game-package-weight pkg)
                                    (package-weight-estimate pkg)))
            :initial-value 0)))

;;; Strategy 1:
;;; If there's a small enough package which is deliverable,
;;; pick it up
(declaim (inline pick-packages we-were-pushed))
(defun pick-packages (bid)
  (let* ((pkgs (packages-at (our-bot-x) (our-bot-y)))
         (deliverable-packages (remove :undeliverable pkgs
                                       :key #'game-package-undeliverable))
         (sorted-packages (sort deliverable-packages #'>
                                :key #'game-package-weight))
         (picked-packages '()))
    (dolist (pkg sorted-packages)
      (when (<= (+ *carrying* (game-package-weight pkg)) *carry-capacity*)
        (push (game-package-id pkg) picked-packages)
        (incf *carrying* (game-package-weight pkg))))

    (send-pick bid picked-packages)))


;; Strategies regarding enemies
(defun we-were-pushed (previous-op)
  ;; Unless the previous was a move along the route, we are not interested
  (when (not previous-op)
    (return-from we-were-pushed nil))
  (debug-print :in-were-pushed :previous-op previous-op)

  (let* ((last-event (robot-events (robot-by-id *own-id*)))
         (move-events
          ;; Filter out X-Y notices
          (loop for x = (pop last-event)
                while x
                if (member x '(:x :y) :test #'eq)
                do (pop last-event)
                ;; We dropped (or picked up something, but that shouldn't
                ;; happen)
                else if (member x '(:d :p) :test #'eq)
                do (return-from we-were-pushed t)
                else collect x)))
    (debug-print :last-event last-event :move-events move-events)
    (not
     (and move-events
          ;; An elaborate way of saying (= (length move-events) 1)
          (null (cdr move-events))
          (eq previous-op (car move-events))))))


;;; SCORING

(declaim (inline score-a-thing))

;; Non-linear scaling to limit the overattractability of
;; larger bases (we can't carry that much stuff anyway)
;;, ie. clip the values to max 1.5 capacity
(defun score-a-thing (weight distance)
  (let ((scaled-weight (min weight (* 1.5 *carry-capacity*))))
    (if (zerop distance) scaled-weight
        (/ scaled-weight distance))))


(defun safe-path-cost (path)
  (or (and path (path-cost path))
      most-positive-fixnum))

;; Handles only unknown bases
;; Since we know the packages anyway, they are also handled
;; at the score-package-location
(defun score-homebases (find-path-fn)
  (loop with best-score = 0
        with best-path = nil
        with best-target = nil
        with x = (our-bot-x)
        with y = (our-bot-y)
        with trimmed-base-list = (maybe-sort-and-snip-bases *homebases* x y)
        for base in trimmed-base-list
        for path = (funcall find-path-fn x y
                            (homebase-x base)
                            (homebase-y base))

        for distance = (safe-path-cost path)
        for weight = (homebase-weight-estimate base)
        for current-score = (score-a-thing weight distance)
        do
        (debug-print :estimating-base base
                     :weight weight
                     :distance distance
                     :current-score current-score)
        if (< best-score current-score)
        do
        (setf best-score current-score
              best-path path
              best-target (list (homebase-x base)
                                (homebase-y base)))
        finally (return (values best-score best-path best-target))))

(defun weigh-package-list (pkg-list)
  (reduce #'+ (remove :undeliverable pkg-list
                      :key #'game-package-undeliverable)
          :key #'package-weight-estimate
          :initial-value 0))

(defun score-package-locations (find-path-fn)
  (loop with best-score = 0
        with best-path = nil
        with best-target = nil
        with x = (our-bot-x)
        with y = (our-bot-y)

        for (pkg-x pkg-y) being the hash-keys of *packages-at* using
            (hash-value pkg-list)
        for path = (funcall find-path-fn x y pkg-x pkg-y)
        for distance = (safe-path-cost path)
        for weight = (weigh-package-list pkg-list)
        for current-score = (score-a-thing weight distance)
        if (and (< best-score current-score) path)
        do
        (setf best-score current-score
              best-path path
              best-target (list pkg-x pkg-y))
        finally (return (values best-score best-path))))

;; When carrying things, find the best home
(defun score-packages-carried (pkgs)
  (loop with best-score = 0
        with best-path = nil
        with best-target = nil
        with x = (our-bot-x)
        with y = (our-bot-y)

        for pkg in pkgs
        for pkg-x = (game-package-to-x pkg)
        for pkg-y = (game-package-to-y pkg)
        for path = (find-path x y pkg-x pkg-y)
        for distance = (safe-path-cost path)
        for weight = (game-package-weight pkg)
        for current-score = (score-a-thing weight distance)
        if (< best-score current-score)
        do
        (setf best-score current-score
              best-path path
              best-target (list pkg-x pkg-y))
        finally (return (values best-score best-path))))

;; Looking for a fight
(defun score-other-players (find-path-fn)
  ;; We go through the loop twice. So shoot me.
  ;; It's bloody 2:40 am, that means 19 hours to go
  ;; and some of the more exotic functions are still
  ;; underway.
  (let* ((players (maphash (lambda (key value) (declare (ignore key)) value)
                                   *robots*))
         (other-players (remove *own-id* players :key #'robot-id)))
    (loop with best-score = 0
          with best-path = nil
          with best-bot = nil
          with x = (our-bot-x)
          with y = (our-bot-y)

          for bot in other-players
          for path = (funcall find-path-fn x y (robot-x bot) (robot-y bot))
          for distance = (safe-path-cost path)
          for weight = (carrying-at-the-moment bot)
          for current-score = (score-a-thing weight distance)
          if (and (< best-score current-score) path)
          do
          (setf best-score current-score
                best-path path
                best-bot bot)
          finally (return (values best-score best-path best-bot)))))

(defun find-target-global (&aux trim)
  ;; Make sure we don't chase anybody if there is anything else
  ;; to do
  (setf *chasing-others* nil)
  (setf *global-target* nil)
  (let* ((path
          (if (zerop *carrying*)
              ;; Not carrying anything, so choose the closest
              ;; package or homebase, depending on which is more
              ;; appealing
              (multiple-value-bind (pkg-score pkg-path pkg-target)
                  (score-package-locations #'find-path)
                (multiple-value-bind (base-score base-path base-target)
                    (score-homebases #'find-path)
                  ;; All out of packages, try to ram someone
                  (cond ((and (zerop pkg-score)
                              (zerop base-score))
                         (setf trim t)
                         (multiple-value-bind (score path target)
                             (score-other-players #'find-path)
                           (declare (ignore score))
                           (setf *chasing-others* target)
                           (setf trim t)
                           path))

                        ((< pkg-score base-score)
                         (setf *global-target* base-target)
                         base-path)

                        (t (setf *global-target* pkg-target)
                           pkg-path))))
              ;; Otherwise find the highest-scoring
              ;; package we are carrying and go there
              (multiple-value-bind (score path target)
                  (score-packages-carried (robot-packages
                                           (robot-by-id *own-id*)))
                (declare (ignore score))
                (setf *global-target* target)
                path)))

         (route (and path (nreverse (path-reversed path)))))
    (if trim
        ;; Don't chase other bots too far
        ;; they will move by the time we get to them
        (subseq route 0 15)
        route)))


;; Verify the strategy
;; If there have been picks or drops in the screen we were intending to go,
;; recompute the strategy
(defun strategy-still-ok-p ()
  (if (null *global-target*)
      ;; Delivering the goods, or chasing someone
      t
      (loop for robot being the hash-values of *robots*
            if (and (= (first *global-target*) (robot-x robot))
                    (= (second *global-target*) (robot-y robot))
                    (not (= *own-id* (robot-id robot)))
                    (or (find :p (robot-events robot))
                        (find :d (robot-events robot))))
            do (return)
            finally (return t))))

;;; Local cases
(defun look-for-local-base ()
  (let* ((x (our-bot-x))
         (y (our-bot-y))
         ;; These are "close enough"
         (interesting-bases
          (remove-if-not (lambda (base) (< (manhattan-heuristic
                                            x y (homebase-x base)
                                            (homebase-y base))
                                           +lookout-window+))
                         *homebases*)))
    (multiple-value-bind (dist path target)
        (loop with closest-target = nil
              with closest-path = nil
              with closest-distance = most-positive-fixnum
              for base in interesting-bases
              for path = (find-path-localized x y (homebase-x base)
                                              (homebase-y base))
              if (and path (< (path-cost path) closest-distance))
              do (setf closest-path path
                       closest-distance (path-cost path)
                       closest-target (list x y))
              finally (return (values closest-distance
                                      closest-path closest-target)))
      (declare (ignore dist))
      (when path
        (setf *global-target* target)
        (nreverse (path-reversed path))))))

;; This doesn't look by distance.
;; Instead it looks by
(defun look-for-local-packages ()
  (let* ((x (our-bot-x))
         (y (our-bot-y))
         pkg
         ;; These are "close enough"
         (interesting-packages
          (loop for i
                from (max 0 (- x +lookout-window+))
                to (min (+ x +lookout-window+) (+ *world-width* 1))
                nconc
                (loop for j
                      from (max 0 (- y +lookout-window+))
                      to (min (+ y +lookout-window+) (+ *world-height* 1))
                      append (packages-at i j))))
         ;; Distribute the packages to two categories
         ;; Again, two passes (where there could theoretically be
         ;; just one, during the topmost loop. So f*cking what.
         (known-packages
          (remove-if-not (lambda (pkg) (and
                                        (< (+ *carrying*
                                              (game-package-weight pkg))
                                           *carry-capacity*)
                                        (eq (game-package-undeliverable pkg) t)))
                         interesting-packages)))

    ;; We do find-path twice for the winning package
    ;; My heart bleeds
    (when known-packages
      (setf pkg (car (remove-if-not
                      (lambda (pkg)
                        (find-path x y (game-package-at-x pkg)
                                   (game-package-at-y pkg)))
                      (sort known-packages #'> :key #'game-package-weight)))))

    (unless pkg
      (setf pkg (car (remove-if-not
                      (lambda (pkg)
                        (find-path x y (game-package-at-x pkg)
                                   (game-package-at-y pkg)))
                      (sort interesting-packages #'>
                            :key #'game-package-interest-factor)))))

      (when pkg
        (nreverse (path-reversed (find-path x y (game-package-at-x pkg)
                                            (game-package-at-y pkg)))))))

;; PANIC CASES -- DO SOMETHING
(defun find-safe-map-square ()
  (let* ((x (our-bot-x))
         (y (our-bot-y))
         (s (cons :s (aref *world-map* x (1- y))))
         (n (cons :n (aref *world-map* x (1+ y))))
         (e (cons :e (aref *world-map* (1+ x) y)))
         (w (cons :w (aref *world-map* (1+ x) y)))
         (directions (list n s e w))

         (safe-place (find #\. directions :key #'cdr))
         (almost-safe (find #\, directions :key #'cdr))
         (wall (find #\# directions :key #'cdr)))
    ;; Try these options
    ;; Completely safe place
    ;; next to water
    ;; bump to a wall
    ;; go west (it will most likely be a suicide, but
    ;; at least it's not a malformed operation)
    (or (car safe-place) (car almost-safe) (car wall) :w)))
